@@ -9,9 +9,16 @@
 package LaTeX::TOM::Parser;
 
 use strict;
+use base qw(
+    LaTeX::TOM::Node
+    LaTeX::TOM::Tree
+);
+use constant true  => 1;
+use constant false => 0;
 
-use base qw(LaTeX::TOM::Node);
-use base qw(LaTeX::TOM::Tree);
+use Carp qw(carp croak);
+
+our $VERSION = '0.02';
 
 # Constructor
 #
@@ -30,17 +37,21 @@ sub new {
 # Set/reset "globals"
 #
 sub _init {
-    my $parser = $_[0];
+    my $parser = shift;
+    my ($parse_errors_fatal, $readinputs, $applymappings) = @_;
 
-    my $parse_errors_fatal = (defined $_[1] ? $_[1] : $parser->{config}{PARSE_ERRORS_FATAL});
-    my $readinputs = (defined $_[2] ? $_[2] : 0);
-    my $applymappings = (defined $_[3] ? $_[3] : 0);
+    my $retrieve_opt_with_default = sub
+    {
+        my ($opt, $default) = @_;
+        return $opt if defined $opt;
+        return $default;
+    };
 
     # set user options
     #
-    $parser->{readinputs} = $readinputs;
-    $parser->{applymappings} = $applymappings;
-    $parser->{PARSE_ERRORS_FATAL} = $parse_errors_fatal;
+    $parser->{readinputs}         = $retrieve_opt_with_default->($readinputs, 0);
+    $parser->{applymappings}      = $retrieve_opt_with_default->($applymappings, 0);
+    $parser->{PARSE_ERRORS_FATAL} = $retrieve_opt_with_default->($parse_errors_fatal, $parser->{config}{PARSE_ERRORS_FATAL});
 
     # init internal stuff
     #
@@ -68,17 +79,13 @@ sub parseFile {
     $parser->{file} = $filename;        # file name member data
     my $tree = {};                      # init output tree
 
-    # read in text from file
+    # read in text from file or bomb out
     #
-    my $text = _readFile($filename);
+    my $text = _readFile($filename, true);
 
-    # do the parse or bomb out
+    # do the parse
     #
-    if ($text) {
-        $tree = $parser->parse($text);
-    } else {
-        die "Could not read file $filename !";
-    }
+    $tree = $parser->parse($text);
 
     return $tree;
 }
@@ -234,7 +241,7 @@ sub _stage2 {
                             {type => 'GROUP',
                              start => $textnode2->{start} - 1,
                              end => $textnode2->{end} + 1,
-                             children => LaTeX::TOM::Tree->new([$textnode2], $parser)
+                             children => LaTeX::TOM::Tree->new([$textnode2], $parser),
                             });
 
                         # splice the new subtree into the old location
@@ -271,10 +278,11 @@ sub _stage2 {
                              start => $textnode2->{start} - 1,
                              end => $textnode3->{end} + 1,
                              children => LaTeX::TOM::Tree->new(
-                                [$textnode2, 
+                                [$textnode2,
                                  @removed,
-                                 $textnode3]),
-                                 $parser});
+                                 $textnode3],
+                                 $parser),
+                            });
 
                         # replace the two original text nodes with the leftover left and
                         # right portions, as well as the group node with everything in
@@ -357,7 +365,7 @@ sub _stage3 {
             # due to the previous parsing stages. if found, the parent node is
             # promoted to a command.
             #
-            if ($text =~ /^\s*\\(\w+)/ && defined $parent && defined $parser->{config}{INNERCMDS}->{$1}) {
+            if ($text =~ /^\s*\\(\w+\*?)/ && defined $parent && $parser->{config}{INNERCMDS}->{$1}) {
                 my $command = $1;
 
                 # if the parent is already a command node, we have to make a new
@@ -397,14 +405,14 @@ sub _stage3 {
                     $parser->{USED_COMMANDS}->{$parent->{command}} = 1;
                 }
 
-                $node->{content} =~ s/^\s*\\(?:\w+)//o;
+                $node->{content} =~ s/^\s*\\(?:\w+\*?)//o;
             }
 
             # outer command (such as \command{parameters}). our regexp checks to
             # see if this text chunk ends in \command, since that would be the case
             # due to the previous parsing stages.
             #
-            if ($text =~ /(?:^|[^\\])(\\\w+(\s*\[.*?\])?)\s*$/os && 
+            if ($text =~ /(?:^|[^\\])(\\\w+\*?(\s*\[.*?\])?)\s*$/os && 
                     defined $tree->{nodes}[$i+1] &&
                     $tree->{nodes}[$i+1]->{type} eq 'GROUP') {
 
@@ -413,10 +421,10 @@ sub _stage3 {
                 #print "found text node [$text] with command tag [$tag]\n";
 
                 # remove the text
-                $node->{content} =~ s/\\\w+\s*(?:\[.*?\])?\s*$//os;
+                $node->{content} =~ s/\\\w+\*?\s*(?:\[.*?\])?\s*$//os;
 
                 # parse it for command and ops
-                $tag =~ /^\\(\w+)\s*(?:\[(.*?)\])?$/os;
+                $tag =~ /^\\(\w+\*?)\s*(?:\[(.*?)\])?$/os;
 
                 my $command = $1;
                 my $opts = $2;
@@ -434,12 +442,12 @@ sub _stage3 {
 
             # recognize braceless commands
             #
-            if ($text =~ /(\\(\w+)[ \t]+(\S+))/gso || $text =~ /(\\(\w+)(\d+))/gso) {
+            if ($text =~ /(\\(\w+\*?)[ \t]+(\S+))/gso || $text =~ /(\\(\w+)(\d+))/gso) {
                 my $all = $1;
                 my $command = $2;
                 my $param = $3;
 
-                if (defined $parser->{config}{BRACELESS}->{$command}) {
+                if ($parser->{config}{BRACELESS}->{$command}) {
                  # warn "found braceless command $command with param $param";
 
                     # get location to split from node text
@@ -466,7 +474,8 @@ sub _stage3 {
                          command => $command,
                          start => $node->{start} + $a,
                          end => $node->{start} + $b,
-                         children => LaTeX::TOM::Tree->new([$newchild]), $parser });
+                         children => LaTeX::TOM::Tree->new([$newchild], $parser),
+                        });
 
                     $parser->{USED_COMMANDS}->{$commandnode->{command}} = 1;
 
@@ -563,9 +572,10 @@ sub _stage4 {
                      end => $end,
                      ostart => $start - length('begin') - length($class) - 2,
                      oend => $end + length('end') + length($class) + 2,
-                     children => LaTeX::TOM::Tree->new([@newarray]), $parser });
+                     children => LaTeX::TOM::Tree->new([@newarray], $parser),
+                    });
 
-                if (defined $parser->{config}{MATHENVS}->{$envnode->{class}}) {
+                if ($parser->{config}{MATHENVS}->{$envnode->{class}}) {
                     $envnode->{math} = 1;
                 }
 
@@ -681,7 +691,8 @@ sub _stage5_r {
                                 ostart => $startpos + $leftpos - length($left) + 1,
                                 end => $startpos + $rightpos,
                                 oend => $startpos + $rightpos + length($right) - 1,
-                                children => LaTeX::TOM::Tree->new([$textnode2]), $parser });
+                                children => LaTeX::TOM::Tree->new([$textnode2], $parser),
+                                });
 
                             splice @{$tree->{nodes}}, $i, 1, $textnode1, $mathnode, $textnode3;
 
@@ -719,7 +730,8 @@ sub _stage5_r {
                                 children => LaTeX::TOM::Tree->new(
                                 [$textnode2,
                                  @remnodes,
-                                 $textnode3]), $parser });
+                                 $textnode3], $parser),
+                                });
 
                             # replace (TEXT_A, ... , TEXT_B) with the mathnode created above
                             splice @{$tree->{nodes}}, $leftidx, 2, $textnode1, $mathnode, $textnode4;
@@ -807,18 +819,20 @@ sub _propegateModes {
             # handle math or plain text forcing envs
             #
             if ($node->{type} eq 'ENVIRONMENT' || $node->{type} eq 'COMMAND') {
-                if (defined $parser->{config}{MATHENVS}->{$node->{class}} || 
-                    defined $parser->{config}{MATHENVS}->{"$node->{class}*"}) {
-
+                if (defined $node->{class} && (
+                    $parser->{config}{MATHENVS}->{$node->{class}} ||
+                    $parser->{config}{MATHENVS}->{"$node->{class}*"})
+                   )
+                {
                     $mathflag = 1;
                     $plaintextflag = 0;
                 }
                 elsif (($node->{type} eq 'COMMAND' && 
-                                (defined $parser->{config}{TEXTENVS}->{$node->{command}} ||
-                                 defined $parser->{config}{TEXTENVS}->{"$node->{command}*"})) ||
+                                ($parser->{config}{TEXTENVS}->{$node->{command}} ||
+                                 $parser->{config}{TEXTENVS}->{"$node->{command}*"})) ||
                              ($node->{type} eq 'ENVIRONMENT' && 
-                                (defined $parser->{config}{TEXTENVS}->{$node->{class}} ||
-                                 defined $parser->{config}{TEXTENVS}{"$node->{command}*"}))
+                                ($parser->{config}{TEXTENVS}->{$node->{class}} ||
+                                 $parser->{config}{TEXTENVS}{"$node->{command}*"}))
                             ) {
 
                     $mathflag = 0;
@@ -986,7 +1000,7 @@ sub _applyMapping {
                  my $idx = index $node->{content}, '\\' . $command, 0;
 
                  # split the text node at that command
-                 my ($leftnode, $rightnode) = splitTextNode($node, $idx, $idx + length($command));
+                 my ($leftnode, $rightnode) = $node->split($idx, $idx + length($command));
 
                  # copy the mapping template
                  my $applied = $mapping->{template}->copy();
@@ -1098,7 +1112,7 @@ sub _applyMappings {
              # make the mapping
              my $mapping = {name => $node->{command},
                 nparams => 0,
-                template => copyTree($node->{children}),
+                template => $node->{children}->copy(),
                 type => 'command'};
 
              next if (!$mapping->{name}); # skip fragged commands
@@ -1156,28 +1170,33 @@ sub _addInputs {
             #print "reading input file $file\n";
 
             # read in contents of file
-            my $contents = _readFile($file);
-            if ($contents eq '' && not $file =~ /\.tex$/) {
+            my $contents;
+            if (-e $file) {
+                $contents = _readFile($file);
+            }
+            elsif ($file !~ /\.tex$/) {
                 $file = "$file.tex";
-                $contents = _readFile($file); 
+                $contents = _readFile($file);
             }
 
             # dump Psfig/TeX files, they aren't useful to us and have
             # nonconforming syntax. Use declaration line as our heuristic.
             #
-            if ($contents =~ /^\%\s*Psfig\/TeX\s*$/m) {
-                $contents = '';
-                warn "ignoring Psfig input '$file'";
+            if (defined $contents
+                     && $contents =~ m!^ \% \s*? Psfig/TeX \s* $!mx
+            ) {
+                undef $contents;
+                carp "ignoring Psfig input `$file'";
             }
 
             # actually do the parse of the sub-content
             #
-            if ($contents) {
+            if (defined $contents) {
                 # parse into a tree
                 my ($subtree,) = $parser->_basicparse($contents, $parser->{PARSE_ERRORS_FATAL});
 
                 # replace \input command node with subtree
-                splice @{$tree->{nodes}}, $i, 1, @$subtree;
+                splice @{$tree->{nodes}}, $i, 1, @{$subtree->{nodes}};
 
                 # step back
                 $i--;
@@ -1194,7 +1213,7 @@ sub _addInputs {
 
                      my $contents = _readFile($file);
 
-                     if ($contents) {
+                     if (defined $contents) {
 
                          my ($subtree,) = $parser->_basicparse($contents, $parser->{PARSE_ERRORS_FATAL});
                          splice @{$tree->{nodes}}, $i, 1, @$subtree;
@@ -1351,7 +1370,7 @@ sub _findbrace {
         $realbrace = 1;
         $index_c = index $text, '}', $pos_c;
 
-        if (substr($text, $index_c - 1, 1) eq ' ') {
+        if (($index_c - 1) >= 0 && substr($text, $index_c - 1, 1) eq ' ') {
             $pos_c = $index_c + 1;
             $index_c = -1;
         }
@@ -1710,17 +1729,19 @@ sub _getTextAndCommentNodes {
 # Read in the contents of a text file on disk. Return in string scalar.
 #
 sub _readFile {
-    my $file = shift || '/dev/stdin';
+    my ($file, $raise_error) = @_;
 
-    my $contents = "";
+    $raise_error ||= false;
 
-    open INFILE, $file or return "";
+    my $opened = open(my $fh, '<', $file);
 
-    while (my $line = <INFILE>) {
-        $contents .= $line;
+    unless ($opened) {
+        croak "Cannot open $file: $!" if $raise_error;
+        return undef;
     }
 
-    close INFILE;
+    my $contents = do { local $/; <$fh> };
+    close($fh);
 
     return $contents;
 }
